# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import os
import torch
import transformers
from transformers import BertModel, XLMRobertaModel

from massive_serve.contriever.src import utils


class Contriever(BertModel):
    def __init__(self, config, pooling="average", **kwargs):
        super().__init__(config, add_pooling_layer=False)
        if not hasattr(config, "pooling"):
            self.config.pooling = pooling

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        output_attentions=None,
        output_hidden_states=None,
        normalize=False,
    ):

        model_output = super().forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )

        last_hidden = model_output["last_hidden_state"]
        last_hidden = last_hidden.masked_fill(~attention_mask[..., None].bool(), 0.0)

        if self.config.pooling == "average":
            emb = last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]
        elif self.config.pooling == "cls":
            emb = last_hidden[:, 0]

        if normalize:
            emb = torch.nn.functional.normalize(emb, dim=-1)
        return emb


class XLMRetriever(XLMRobertaModel):
    def __init__(self, config, pooling="average", **kwargs):
        super().__init__(config, add_pooling_layer=False)
        if not hasattr(config, "pooling"):
            self.config.pooling = pooling

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        output_attentions=None,
        output_hidden_states=None,
        normalize=False,
    ):

        model_output = super().forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )

        last_hidden = model_output["last_hidden_state"]
        last_hidden = last_hidden.masked_fill(~attention_mask[..., None].bool(), 0.0)
        if self.config.pooling == "average":
            emb = last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]
        elif self.config.pooling == "cls":
            emb = last_hidden[:, 0]
        if normalize:
            emb = torch.nn.functional.normalize(emb, dim=-1)
        return emb


def load_retriever(model_path, pooling="average", random_init=False):
    # try: check if model exists locally
    path = os.path.join(model_path, "checkpoint.pth")
    if os.path.exists(path):
        pretrained_dict = torch.load(path, map_location="cpu")
        opt = pretrained_dict["opt"]
        if hasattr(opt, "retriever_model_id"):
            retriever_model_id = opt.retriever_model_id
        else:
            # retriever_model_id = "bert-base-uncased"
            retriever_model_id = "bert-base-multilingual-cased"
        tokenizer = utils.load_hf(transformers.AutoTokenizer, retriever_model_id)
        cfg = utils.load_hf(transformers.AutoConfig, retriever_model_id)
        if "xlm" in retriever_model_id:
            model_class = XLMRetriever
        else:
            model_class = Contriever
        retriever = model_class(cfg)
        pretrained_dict = pretrained_dict["model"]

        if any("encoder_q." in key for key in pretrained_dict.keys()):  # test if model is defined with moco class
            pretrained_dict = {k.replace("encoder_q.", ""): v for k, v in pretrained_dict.items() if "encoder_q." in k}
        elif any("encoder." in key for key in pretrained_dict.keys()):  # test if model is defined with inbatch class
            pretrained_dict = {k.replace("encoder.", ""): v for k, v in pretrained_dict.items() if "encoder." in k}
        retriever.load_state_dict(pretrained_dict, strict=False)
    else:
        retriever_model_id = model_path
        if "xlm" in retriever_model_id:
            model_class = XLMRetriever
        else:
            model_class = Contriever
        cfg = utils.load_hf(transformers.AutoConfig, model_path)
        tokenizer = utils.load_hf(transformers.AutoTokenizer, model_path)
        retriever = utils.load_hf(model_class, model_path)

    return retriever, tokenizer, retriever_model_id



import os
import json
import csv
import pickle
import gzip
import logging
import pdb

from massive_serve.src.indicies.index_utils import get_passage_pos_ids


############################## Training ##############################
def fast_load_jsonl_shard(passages_dir, shard_index, return_all_passages=True):
    """
    This function is designed to handle large datasets by only loading the specific portion of data (shard) that 
    corresponds to the given shard index.

    Shards are determined by dividing the total size of all files in the directory evenly by `num_shards`. 
    This function reads only the data portion of the `shard_index` shard, chunks the text from each line 
    based on `chunk_sz`, and appends each chunk to a list with an incremental ID.
    """
    use_passage_pos_id_map = True
    
    if not return_all_passages:
        assert use_passage_pos_id_map, f"You must set `use_passage_pos_id_map=True` to enable efficient passage loading!"

    # Check the existance of processed data
    passage_shard_save_path = os.path.join(passages_dir, f'raw_passages-{shard_index}-of-{num_shards}.jsonl')
    pos_map_save_path = os.path.join(passages_dir, 'passage_pos_id_map.pkl')

    if not return_all_passages:
        with open(pos_map_save_path, 'rb') as f:
            passage_pos_ids = pickle.load(f)
        return  passage_pos_ids
    
    elif os.path.exists(passage_shard_save_path):
        passages = []
        with open(passage_shard_save_path, 'r') as fin:
            for line in fin:
                passages.append(json.loads(line))
        return passages



import os
import sys
import subprocess
import click
from .data_utils import prepare_for_upload, prepare_after_download

@click.group()
def cli():
    """Massive Serve CLI"""
    pass

@cli.command(name='upload-data')
@click.option('--chunk_size_gb', type=float, default=40, help='Maximum size of each chunk in GB')
@click.option('--domain_name', type=str, default='dpr_wiki_contriever', help='Domain name')
def upload_data(chunk_size_gb, domain_name):
    """Upload data to Hugging Face, automatically splitting large files"""
    # Set datastore path to `~` if it is not already set
    env = os.environ.copy()
    if 'DATASTORE_PATH' not in env:
        env['DATASTORE_PATH'] = os.path.expanduser('~')
    
    data_dir = os.path.join(os.path.expanduser(env['DATASTORE_PATH']), domain_name)
    
    # Split large files if necessary
    split_files = prepare_for_upload(data_dir, chunk_size_gb)
    if split_files:
        print(f"Split {len(split_files)} files: {split_files}")
    
    # Upload to Hugging Face
    subprocess.run(['huggingface-cli', 'upload', f'rulins/massive_serve_{domain_name}', data_dir, '--repo-type', 'dataset'])

@cli.command(name='serve')
@click.option('--domain_name', type=str, default='dpr_wiki_contriever', help='Domain name')
def serve(domain_name):
    """Run the worker node"""
    # Set the domain name
    os.environ['MASSIVE_SERVE_DOMAIN_NAME'] = domain_name
    
    # Set datastore path to `~` if it is not already set
    env = os.environ.copy()
    if 'DATASTORE_PATH' not in env:
        datastore_path = input(f"Please enter a path to save the downloaded index (default: {os.path.expanduser('~')}): ").strip()
        env['DATASTORE_PATH'] = datastore_path if datastore_path else os.path.expanduser('~')
    
    # Add package root to PYTHONPATH
    package_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    if package_root not in sys.path:
        sys.path.insert(0, package_root)
    
    # Download the wiki index dataset
    save_path = os.path.join(os.path.expanduser(env['DATASTORE_PATH']), domain_name)
    subprocess.run(['huggingface-cli', 'download', f'rulins/massive_serve_{domain_name}', '--repo-type', 'dataset', '--local-dir', save_path])
    
    # Combine any split files
    print("Combining split files...")
    prepare_after_download(save_path)
    
    # Verify that the index file exists
    index_dir = os.path.join(save_path, 'index')
    index_files = [f for f in os.listdir(index_dir) if f.endswith('.faiss')]
    if len(index_files) != 1:
        raise FileNotFoundError(f"Expected exactly one .faiss file in {index_dir}, found {len(index_files)}")
    index_path = os.path.join(index_dir, index_files[0])
    if not os.path.exists(index_path):
        raise FileNotFoundError(f"Index file not found at {index_path} after combining split files")
    
    print(f"Starting {domain_name} server...")
    # Run the worker node script
    from .api.serve import main as serve_main
    serve_main()

if __name__ == '__main__':
    cli() 


import os
import json
import traceback
import datetime
import socket
import getpass
from flask import Flask, jsonify, request
from flask_cors import CORS
import threading
import queue

from massive_serve.api.api_index import get_datastore

# ANSI color codes
CYAN = '\033[36m'
GREEN = '\033[32m'
YELLOW = '\033[33m'
WHITE = '\033[37m'
RESET = '\033[0m'


class DatastoreConfig():
    def __init__(self,):
        super().__init__()
        self.data_root = os.environ.get('DATASTORE_PATH', os.path.expanduser('~'))
        self.domain_name = os.environ.get('MASSIVE_SERVE_DOMAIN_NAME', 'demo')
        self.config = json.load(open(os.path.join(self.data_root, self.domain_name, 'config.json')))
        assert self.config['domain_name'] == self.domain_name, f"Domain name in config.json ({self.config['domain_name']}) does not match the domain name in the environment variable ({self.domain_name})"
        self.query_encoder = self.config['query_encoder']
        self.query_tokenizer = self.config['query_tokenizer']
        self.index_type = self.config['index_type']
        self.per_gpu_batch_size = self.config['per_gpu_batch_size']
        self.question_maxlength = self.config['question_maxlength']
        self.nprobe = self.config['nprobe']

ds_cfg = DatastoreConfig()
        

app = Flask(__name__)
CORS(app)


class Item:
    def __init__(self, query=None, query_embed=None, domains=ds_cfg.domain_name, n_docs=1, nprobe=None) -> None:
        self.query = query
        self.query_embed = query_embed
        self.domains = domains
        self.n_docs = n_docs
        self.nprobe = nprobe
        self.searched_results = None
    
    def get_dict(self,):
        dict_item = {
            'query': self.query,
            'query_embed': self.query_embed,
            'domains': self.domains,
            'n_docs': self.n_docs,
            'nprobe': self.nprobe,
            'searched_results': self.searched_results,
        }
        return dict_item


class SearchQueue:
    def __init__(self, log_queries=False):
        self.queue = queue.Queue()
        self.lock = threading.Lock()
        self.current_search = None
        self.datastore = get_datastore(ds_cfg)

        self.log_queries = log_queries
        self.query_log = 'cached_queries.jsonl'
    
    def search(self, item):
        with self.lock:
            if self.current_search is None:
                self.current_search = item
                if item.nprobe is not None:
                    self.datastore.index.nprobe = item.nprobe
                results = self.datastore.search(item.query, item.n_docs)
                self.current_search = None
                return results
            else:
                future = threading.Event()
                self.queue.put((item, future))
                future.wait()
                return item.searched_results
    
    def process_queue(self):
        while True:
            item, future = self.queue.get()
            with self.lock:
                self.current_search = item
                item.searched_results = self.datastore.search(item)
                self.current_search = None
            future.set()
            self.queue.task_done()

search_queue = SearchQueue()
threading.Thread(target=search_queue.process_queue, daemon=True).start()

@app.route('/search', methods=['POST'])
def search():
    try:
        item = Item(
            query=request.json['query'],
            n_docs=request.json.get('n_docs', 1),
            nprobe=request.json.get('nprobe', None),
            domains=ds_cfg.domain_name,
        )
        # Perform the search synchronously with 60s timeout
        timer = threading.Timer(60.0, lambda: (_ for _ in ()).throw(TimeoutError('Search timed out after 60 seconds')))
        timer.start()
        try:
            results = search_queue.search(item)
            timer.cancel()
            print(results)
            return jsonify({
                "message": f"Search completed for '{item.query}' from {item.domains}",
                "query": item.query,
                "n_docs": item.n_docs,
                "nprobe": item.nprobe,
                "results": results,
            }), 200
        except TimeoutError as e:
            timer.cancel()
            return jsonify({
                "message": str(e),
                "query": item.query,
                "error": "timeout"
            }), 408
    except Exception as e:
        tb_lines = traceback.format_exception(e.__class__, e, e.__traceback__)
        error_message = f"An error occurred: {str(e)}\n{''.join(tb_lines)}"
        return jsonify({"message": error_message}), 500

@app.route('/current_search')
def current_search():
    with search_queue.lock:
        current = search_queue.current_search
        if current:
            return jsonify({
                "current_search": current.query,
                "domains": current.domains,
                "n_docs": current.n_docs,
            }), 200
        else:
            return jsonify({"message": "No search currently in progress"}), 200

@app.route('/queue_size')
def queue_size():
    size = search_queue.queue.qsize()
    return jsonify({"queue_size": size}), 200

@app.route("/")
def home():
    return jsonify("Hello! What you are looking for?")


def find_free_port():
    with socket.socket() as s:
        s.bind(("", 0))  # Bind to a free port provided by the host.
        return s.getsockname()[1]  # Return the port number assigned.


def main():
    port = find_free_port()
    server_id = socket.gethostname()
    domain_name = ds_cfg.domain_name
    serve_info = {'server_id': server_id, 'port': port}
    username = os.environ.get('USER') or os.environ.get('USERNAME') or getpass.getuser()
    endpoint = f'{username}@{server_id}:{port}/search'  # use actual username
    
    # Create a beautiful banner
    banner = f"""
{CYAN}╔════════════════════════════════════════════════════════════╗
║{GREEN}                    MASSIVE SERVE SERVER                    {CYAN}║
╠════════════════════════════════════════════════════════════╣
║{YELLOW} Domain: {WHITE}{domain_name:<45}{CYAN}║
║{YELLOW} Server: {WHITE}{server_id:<45}{CYAN}║
║{YELLOW} Port:   {WHITE}{port:<45}{CYAN}║
║{YELLOW} Endpoint: {WHITE}{endpoint:<41}{CYAN}║
╚════════════════════════════════════════════════════════════╝{RESET}
"""
    print(banner)
    
    # Print test request with colors
    test_request = f"""
{GREEN}Test your server with this curl command:\n{RESET}
{CYAN}curl -X POST {endpoint} -H "Content-Type: application/json" -d '{YELLOW}{{"query": "Tell me more about the stories of Einstein.", "n_docs": 1, "domains": "{domain_name}"}}{CYAN}'{RESET}
"""
    print(test_request)
    
    app.run(host='0.0.0.0', port=port)


if __name__ == '__main__':
    main()


import os
import json
import random
import logging
import pickle
import time
import glob
from tqdm import tqdm
import pdb
import re

import faiss
import numpy as np
import torch


os.environ["TOKENIZERS_PARALLELISM"] = "true"

device = 'cuda' if torch.cuda.is_available()  else 'cpu'


class IVFPQIndexer(object):

    def __init__(self, 
                embed_paths,
                index_path,
                meta_file,
                trained_index_path,
                passage_dir=None,
                pos_map_save_path=None,
                sample_train_size=1000000,
                prev_index_path=None,
                dimension=768,
                dtype=np.float16,
                ncentroids=4096,
                probe=2048,
                num_keys_to_add_at_a_time=1000000,
                DSTORE_SIZE_BATCH=51200000,
                n_subquantizers=16,
                code_size=8,
                ):
    
        self.embed_paths = embed_paths  # list of paths where saved the embedding of all shards
        self.index_path = index_path  # path to store the final index
        self.meta_file = meta_file  # path to save the index id to db id map
        self.prev_index_path = prev_index_path  # add new data to it instead of training new clusters
        self.trained_index_path = trained_index_path  # path to save the trained index
        self.passage_dir = passage_dir
        self.pos_map_save_path = pos_map_save_path
        self.cuda = False

        self.sample_size = sample_train_size
        self.dimension = dimension
        self.ncentroids = ncentroids
        self.probe = probe
        self.num_keys_to_add_at_a_time = num_keys_to_add_at_a_time
        self.n_subquantizers = n_subquantizers
        self.code_size = code_size

        print("Loading index...")
        self.index = faiss.read_index(index_path)
        self.index_id_to_db_id = self.load_index_id_to_db_id()
        self.index.nprobe = self.probe
            
        if self.pos_map_save_path is not None:
            self.psg_pos_id_map = self.load_psg_pos_id_map()

    def load_index_id_to_db_id(self,):
        with open(self.meta_file, "rb") as reader:
            index_id_to_db_id = pickle.load(reader)
        return index_id_to_db_id
    
    def load_embeds(self, shard_id=None):
        all_ids, all_embeds = [], []
        offset = 0
        for embed_path in self.embed_paths:
            loaded_shard_id = int(re.search(r'_(\d+).pkl$', embed_path).group(1))
            if shard_id is not None and loaded_shard_id != shard_id:
                continue
            print(f"Loading pickle embedding from {embed_path}...")
            with open(embed_path, "rb") as fin:
                ids, embeddings = pickle.load(fin)
            all_ids.extend([i + offset for i in ids])
            all_embeds.extend(embeddings)
            offset += len(ids)
        all_embeds = np.stack(all_embeds).astype(np.float32)
        datastore_size = len(all_ids)
        return all_embeds
        
    def get_embs(self, indices=None, shard_id=None):
        if indices is not None:
            embs = self.embs[indices]
        elif shard_id is not None:
            embs = self.load_embeds(shard_id)
        return embs

    def get_knn_scores(self, query_emb, indices):
        assert self.embeds is not None
        embs = self.get_embs(indices) # [batch_size, k, dimension]
        scores = - np.sqrt(np.sum((np.expand_dims(query_emb, 1)-embs)**2, -1)) # [batch_size, k]
        return scores

    def _sample_and_train_index(self,):
        print(f"Sampling {self.sample_size} examples from {len(self.embed_paths)} files...")
        per_shard_sample_size = self.sample_size // len(self.embed_paths)
        all_sampled_embs = []
        for embed_path in self.embed_paths:
            print(f"Loading pickle embedding from {embed_path}...")
            with open(embed_path, "rb") as fin:
                _, embeddings = pickle.load(fin)
            shard_size = len(embeddings)
            print(f"Finished loading, sampling {per_shard_sample_size} from {shard_size} for training...")
            random_samples = np.random.choice(np.arange(shard_size), size=[min(per_shard_sample_size, shard_size)], replace=False)
            sampled_embs = embeddings[random_samples]
            all_sampled_embs.extend(sampled_embs)
        all_sampled_embs = np.stack(all_sampled_embs).astype(np.float32)
        
        print ("Training index...")
        start_time = time.time()
        self._train_index(all_sampled_embs, self.trained_index_path)
        print ("Finish training (%ds)" % (time.time()-start_time))

    def _train_index(self, sampled_embs, trained_index_path):
        quantizer = faiss.IndexFlatIP(self.dimension)
        start_index = faiss.IndexIVFPQ(quantizer,
                                       self.dimension,
                                       self.ncentroids,
                                       self.n_subquantizers,
                                       self.code_size,
                                       faiss.METRIC_INNER_PRODUCT
                                       )
        start_index.nprobe = self.probe
        np.random.seed(1)

        if self.cuda:
            # Convert to GPU index
            res = faiss.StandardGpuResources()
            co = faiss.GpuClonerOptions()
            co.useFloat16 = True
            gpu_index = faiss.index_cpu_to_gpu(res, 0, start_index, co)
            gpu_index.verbose = False

            # Train on GPU and back to CPU
            gpu_index.train(sampled_embs)
            start_index = faiss.index_gpu_to_cpu(gpu_index)
        else:
            # Faiss does not handle adding keys in fp16 as of writing this.
            start_index.train(sampled_embs)
        faiss.write_index(start_index, trained_index_path)

    def _add_keys(self, index_path, trained_index_path):
        index = faiss.read_index(trained_index_path)
        assert index.is_trained and index.ntotal == 0
        
        start_time = time.time()
        # NOTE: the shard id is a absolute id defined in the name
        for embed_path in self.embed_paths:
            filename = os.path.basename(embed_path)
            match = re.search(r"passages_(\d+)\.pkl", filename)
            shard_id = int(match.group(1))
                
            to_add = self.get_embs(shard_id=shard_id).copy()
            index.add(to_add)
            ids_toadd = [[shard_id, chunk_id] for chunk_id in range(len(to_add))]  #TODO: check len(to_add) is correct usage
            self.index_id_to_db_id.extend(ids_toadd)
            print ('Added %d / %d shards, (%d min)' % (shard_id+1, len(self.embed_paths), (time.time()-start_time)/60))
        
        faiss.write_index(index, index_path)
        with open(self.meta_file, 'wb') as fout:
            pickle.dump(self.index_id_to_db_id, fout)
        print ('Adding took {} s'.format(time.time() - start_time))
        return index
    
    def build_passage_pos_id_map(self, ):
        convert_pkl_to_jsonl(self.passage_dir)
        passage_pos_ids = get_passage_pos_ids(self.passage_dir, self.pos_map_save_path)
        return passage_pos_ids

    def load_psg_pos_id_map(self,):
        if os.path.exists(self.pos_map_save_path):
            with open(self.pos_map_save_path, 'rb') as f:
                psg_pos_id_map = pickle.load(f)
        else:
            psg_pos_id_map = self.build_passage_pos_id_map()
        return psg_pos_id_map
    
    def _id2psg(self, shard_id, chunk_id):
        filename, position = self.psg_pos_id_map[shard_id][chunk_id]
        with open(filename, 'r') as file:
            file.seek(position)
            line = file.readline()
        return json.loads(line)
    
    def _get_passage(self, index_id):
        shard_id, chunk_id = self.index_id_to_db_id[index_id]
        return self._id2psg(shard_id, chunk_id)
    
    def get_retrieved_passages(self, all_indices):
        passages, db_ids = [], []
        for query_indices in all_indices:
            passages_per_query = [self._get_passage(int(index_id))["text"] for index_id in query_indices]
            db_ids_per_query = [self.index_id_to_db_id[int(index_id)] for index_id in query_indices]
            passages.append(passages_per_query)
            db_ids.append(db_ids_per_query)
        return passages, db_ids
    
    def search(self, query_embs, k=4096):
        all_scores, all_indices = self.index.search(query_embs.astype(np.float32), k)
        all_passages, db_ids = self.get_retrieved_passages(all_indices)
        return all_scores.tolist(), all_passages, db_ids
        


def test_build_multi_shard_dpr_wiki():
    embed_dir = '/fsx-comem/rulin/data/truth_teller/scaling_out/embeddings/facebook/contriever-msmarco/dpr_wiki/1-shards'
    embed_paths = [os.path.join(embed_dir, filename) for filename in os.listdir(embed_dir) if filename.endswith('.pkl')]
    sample_train_size = 6000000
    projection_size = 768
    ncentroids = 4096
    n_subquantizers = 16
    code_size = 8
    formatted_index_name = f"index_ivf_pq_ip.{sample_train_size}.{projection_size}.{ncentroids}.faiss"
    index_dir = '/fsx-comem/rulin/data/truth_teller/scaling_out/embeddings/facebook/contriever-msmarco/dpr_wiki/1-shards/index_ivf_pq/'
    os.makedirs(index_dir, exist_ok=True)
    index_path = os.path.join(index_dir, formatted_index_name)
    meta_file = os.path.join(index_dir, formatted_index_name+'.meta')
    trained_index_path = os.path.join(index_dir, formatted_index_name+'.trained')
    pos_map_save_path = os.path.join(index_dir, 'passage_pos_id_map.pkl')
    passage_dir = '/fsx-comem/rulin/data/truth_teller/scaling_out/passages/dpr_wiki/1-shards'
    index = IVFPQIndexer(
        embed_paths,
        index_path,
        meta_file,
        trained_index_path,
        passage_dir=passage_dir,
        pos_map_save_path=pos_map_save_path,
        sample_train_size=sample_train_size,
        dimension=projection_size,
        ncentroids=ncentroids,
        n_subquantizers=n_subquantizers,
        code_size=code_size,
        )


def test_build_single_shard_dpr_wiki():
    embed_dir = '/checkpoint/amaia/explore/comem/data/scaling_out/embeddings/facebook/dragon-plus-context-encoder/dpr_wiki/8-shards'
    embed_paths = [os.path.join(embed_dir, filename) for filename in os.listdir(embed_dir) if filename.endswith('.pkl')]
    sample_train_size = 6000000
    projection_size = 768
    ncentroids = 4096
    n_subquantizers = 16
    code_size = 8
    formatted_index_name = f"index_ivf_pq_ip.{sample_train_size}.{projection_size}.{ncentroids}.faiss"
    index_dir = '/checkpoint/amaia/explore/comem/data/scaling_out/embeddings/facebook/dragon-plus-context-encoder/dpr_wiki/8-shards/index_ivf_pq/'
    os.makedirs(index_dir, exist_ok=True)
    index_path = os.path.join(index_dir, formatted_index_name)
    meta_file = os.path.join(index_dir, formatted_index_name+'.meta')
    trained_index_path = os.path.join(index_dir, formatted_index_name+'.trained')
    pos_map_save_path = os.path.join(index_dir, 'passage_pos_id_map.pkl')
    passage_dir = '/fsx-comem/rulin/data/truth_teller/scaling_out/passages/dpr_wiki/8-shards'
    index = IVFPQIndexer(
        embed_paths,
        index_path,
        meta_file,
        trained_index_path,
        passage_dir=passage_dir,
        pos_map_save_path=pos_map_save_path,
        sample_train_size=sample_train_size,
        dimension=projection_size,
        ncentroids=ncentroids,
        n_subquantizers=n_subquantizers,
        code_size=code_size,
        )

if __name__ == '__main__':
    test_build_multi_shard_dpr_wiki()


import os
import re
import json
import time
import pickle
import faiss
import numpy as np
import torch


os.environ["TOKENIZERS_PARALLELISM"] = "true"

device = 'cuda' if torch.cuda.is_available()  else 'cpu'


class FlatIndexer(object):

    def __init__(self, 
                embed_paths=None,
                index_path=None,
                meta_file=None,
                passage_dir=None,
                pos_map_save_path=None,
                dimension=768,
                ):
    
        self.embed_paths = embed_paths
        self.index_path = index_path  # path to store the final index
        self.meta_file = meta_file  # path to save the index id to db id map
        self.passage_dir = passage_dir
        self.pos_map_save_path = pos_map_save_path
        self.dimension=dimension
        self.cuda = False

        if os.path.exists(index_path) and os.path.exists(self.meta_file):
            print("Loading index...")
            self.index = faiss.read_index(index_path)
            self.index_id_to_db_id = self.load_index_id_to_db_id()
        else:
            self.index = faiss.IndexFlatIP(dimension)
            self.index_id_to_db_id = []
            print ("Building index...")
            self._build_index()
        
        if self.pos_map_save_path is not None:
            self.psg_pos_id_map = self.load_psg_pos_id_map()

    def _build_index(self,):
        start_time = time.time()
        for embed_path in self.embed_paths:
            filename = os.path.basename(embed_path)
            match = re.search(r"passages_(\d+)\.pkl", filename)
            shard_id = int(match.group(1))
                
            to_add = self.get_embs(shard_id=shard_id).copy()
            self.index.add(to_add)
            ids_toadd = [[shard_id, chunk_id] for chunk_id in range(len(to_add))]  #TODO: check len(to_add) is correct usage
            self.index_id_to_db_id.extend(ids_toadd)
            print ('Added %d / %d shards, (%d min)' % (shard_id+1, len(self.embed_paths), (time.time()-start_time)/60))
        
        faiss.write_index(self.index, self.index_path)
        with open(self.meta_file, 'wb') as fout:
            pickle.dump(self.index_id_to_db_id, fout)
        print ('Adding took {} s'.format(time.time() - start_time))
        
        print(f'Total data indexed {len(self.index_id_to_db_id)}')
        faiss.write_index(self.index, self.index_path)
        with open(self.meta_file, mode='wb') as f:
            pickle.dump(self.index_id_to_db_id, f)
        
    def load_embeds(self, shard_id=None):
        all_ids, all_embeds = [], []
        offset = 0
        for embed_path in self.embed_paths:
            loaded_shard_id = int(re.search(r'_(\d+).pkl$', embed_path).group(1))
            if shard_id is not None and loaded_shard_id != shard_id:
                continue
            print(f"Loading pickle embedding from {embed_path}...")
            with open(embed_path, "rb") as fin:
                ids, embeddings = pickle.load(fin)
            all_ids.extend([i + offset for i in ids])
            all_embeds.extend(embeddings)
            offset += len(ids)
        all_embeds = np.stack(all_embeds).astype(np.float32)
        datastore_size = len(all_ids)
        return all_embeds

    def get_embs(self, indices=None, shard_id=None):
        if indices is not None:
            embs = self.embs[indices]
        elif shard_id is not None:
            embs = self.load_embeds(shard_id)
        return embs
    
    def load_index_id_to_db_id(self,):
        with open(self.meta_file, "rb") as reader:
            index_id_to_db_id = pickle.load(reader)
        return index_id_to_db_id
    
    def build_passage_pos_id_map(self, ):
        convert_pkl_to_jsonl(self.passage_dir)
        passage_pos_ids = get_passage_pos_ids(self.passage_dir, self.pos_map_save_path)
        return passage_pos_ids

    def load_psg_pos_id_map(self,):
        if os.path.exists(self.pos_map_save_path):
            with open(self.pos_map_save_path, 'rb') as f:
                psg_pos_id_map = pickle.load(f)
        else:
            psg_pos_id_map = self.build_passage_pos_id_map()
        return psg_pos_id_map
    
    def _id2psg(self, shard_id, chunk_id):
        filename, position = self.psg_pos_id_map[shard_id][chunk_id]
        with open(filename, 'r') as file:
            file.seek(position)
            line = file.readline()
        return json.loads(line)
    
    def _get_passage(self, index_id):
        try:
            shard_id, chunk_id = self.index_id_to_db_id[index_id]
        except:
            shard_id, chunk_id = 0, self.index_id_to_db_id[index_id]
        return self._id2psg(shard_id, chunk_id)
    
    def get_retrieved_passages(self, all_indices):
        passages, db_ids = [], []
        for query_indices in all_indices:
            passages_per_query = [self._get_passage(int(index_id))["text"] for index_id in query_indices]
            db_ids_per_query = [self.index_id_to_db_id[int(index_id)] for index_id in query_indices]
            passages.append(passages_per_query)
            db_ids.append(db_ids_per_query)
        return passages, db_ids
    
    def search(self, query_embs, k=4096):
        all_scores, all_indices = self.index.search(query_embs.astype(np.float32), k)
        all_passages, db_ids = self.get_retrieved_passages(all_indices)
        return all_scores.tolist(), all_passages, db_ids



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import os
import sys
import logging
import torch
import errno
from typing import Union, Tuple, List, Dict
from collections import defaultdict

from massive_serve.contriever.src import dist_utils

Number = Union[float, int]

logger = logging.getLogger(__name__)


def init_logger(args, stdout_only=False):
    if torch.distributed.is_initialized():
        torch.distributed.barrier()
    stdout_handler = logging.StreamHandler(sys.stdout)
    handlers = [stdout_handler]
    if not stdout_only:
        file_handler = logging.FileHandler(filename=os.path.join(args.output_dir, "run.log"))
        handlers.append(file_handler)
    logging.basicConfig(
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO if dist_utils.is_main() else logging.WARN,
        format="[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s",
        handlers=handlers,
    )
    return logger


def symlink_force(target, link_name):
    try:
        os.symlink(target, link_name)
    except OSError as e:
        if e.errno == errno.EEXIST:
            os.remove(link_name)
            os.symlink(target, link_name)
        else:
            raise e


def save(model, optimizer, scheduler, step, opt, dir_path, name):
    model_to_save = model.module if hasattr(model, "module") else model
    path = os.path.join(dir_path, "checkpoint")
    epoch_path = os.path.join(path, name)  # "step-%s" % step)
    os.makedirs(epoch_path, exist_ok=True)
    cp = os.path.join(path, "latest")
    fp = os.path.join(epoch_path, "checkpoint.pth")
    checkpoint = {
        "step": step,
        "model": model_to_save.state_dict(),
        "optimizer": optimizer.state_dict(),
        "scheduler": scheduler.state_dict(),
        "opt": opt,
    }
    torch.save(checkpoint, fp)
    symlink_force(epoch_path, cp)
    if not name == "lastlog":
        logger.info(f"Saving model to {epoch_path}")


def load(model_class, dir_path, opt, reset_params=False):
    epoch_path = os.path.realpath(dir_path)
    checkpoint_path = os.path.join(epoch_path, "checkpoint.pth")
    logger.info(f"loading checkpoint {checkpoint_path}")
    checkpoint = torch.load(checkpoint_path, map_location="cpu")
    opt_checkpoint = checkpoint["opt"]
    state_dict = checkpoint["model"]

    model = model_class(opt_checkpoint)
    model.load_state_dict(state_dict, strict=True)
    model = model.cuda()
    step = checkpoint["step"]
    if not reset_params:
        optimizer, scheduler = set_optim(opt_checkpoint, model)
        scheduler.load_state_dict(checkpoint["scheduler"])
        optimizer.load_state_dict(checkpoint["optimizer"])
    else:
        optimizer, scheduler = set_optim(opt, model)

    return model, optimizer, scheduler, opt_checkpoint, step


############ OPTIM


class WarmupLinearScheduler(torch.optim.lr_scheduler.LambdaLR):
    def __init__(self, optimizer, warmup, total, ratio, last_epoch=-1):
        self.warmup = warmup
        self.total = total
        self.ratio = ratio
        super(WarmupLinearScheduler, self).__init__(optimizer, self.lr_lambda, last_epoch=last_epoch)

    def lr_lambda(self, step):
        if step < self.warmup:
            return (1 - self.ratio) * step / float(max(1, self.warmup))

        return max(
            0.0,
            1.0 + (self.ratio - 1) * (step - self.warmup) / float(max(1.0, self.total - self.warmup)),
        )


class CosineScheduler(torch.optim.lr_scheduler.LambdaLR):
    def __init__(self, optimizer, warmup, total, ratio=0.1, last_epoch=-1):
        self.warmup = warmup
        self.total = total
        self.ratio = ratio
        super(CosineScheduler, self).__init__(optimizer, self.lr_lambda, last_epoch=last_epoch)

    def lr_lambda(self, step):
        if step < self.warmup:
            return float(step) / self.warmup
        s = float(step - self.warmup) / (self.total - self.warmup)
        return self.ratio + (1.0 - self.ratio) * math.cos(0.5 * math.pi * s)


def set_optim(opt, model):
    if opt.optim == "adamw":
        optimizer = torch.optim.AdamW(
            model.parameters(), lr=opt.lr, betas=(opt.beta1, opt.beta2), eps=opt.eps, weight_decay=opt.weight_decay
        )
    else:
        raise NotImplementedError("optimizer class not implemented")

    scheduler_args = {
        "warmup": opt.warmup_steps,
        "total": opt.total_steps,
        "ratio": opt.lr_min_ratio,
    }
    if opt.scheduler == "linear":
        scheduler_class = WarmupLinearScheduler
    elif opt.scheduler == "cosine":
        scheduler_class = CosineScheduler
    else:
        raise ValueError
    scheduler = scheduler_class(optimizer, **scheduler_args)
    return optimizer, scheduler


def get_parameters(net, verbose=False):
    num_params = 0
    for param in net.parameters():
        num_params += param.numel()
    message = "[Network] Total number of parameters : %.6f M" % (num_params / 1e6)
    return message


class WeightedAvgStats:
    """provides an average over a bunch of stats"""

    def __init__(self):
        self.raw_stats: Dict[str, float] = defaultdict(float)
        self.total_weights: Dict[str, float] = defaultdict(float)

    def update(self, vals: Dict[str, Tuple[Number, Number]]) -> None:
        for key, (value, weight) in vals.items():
            self.raw_stats[key] += value * weight
            self.total_weights[key] += weight

    @property
    def stats(self) -> Dict[str, float]:
        return {x: self.raw_stats[x] / self.total_weights[x] for x in self.raw_stats.keys()}

    @property
    def tuple_stats(self) -> Dict[str, Tuple[float, float]]:
        return {x: (self.raw_stats[x] / self.total_weights[x], self.total_weights[x]) for x in self.raw_stats.keys()}

    def reset(self) -> None:
        self.raw_stats = defaultdict(float)
        self.total_weights = defaultdict(float)

    @property
    def average_stats(self) -> Dict[str, float]:
        keys = sorted(self.raw_stats.keys())
        if torch.distributed.is_initialized():
            torch.distributed.broadcast_object_list(keys, src=0)
        global_dict = {}
        for k in keys:
            if not k in self.total_weights:
                v = 0.0
            else:
                v = self.raw_stats[k] / self.total_weights[k]
            v, _ = dist_utils.weighted_average(v, self.total_weights[k])
            global_dict[k] = v
        return global_dict


def load_hf(object_class, model_name):
    try:
        obj = object_class.from_pretrained(model_name, local_files_only=True)
    except:
        obj = object_class.from_pretrained(model_name, local_files_only=False)
    return obj


def init_tb_logger(output_dir):
    try:
        from torch.utils import tensorboard

        if dist_utils.is_main():
            tb_logger = tensorboard.SummaryWriter(output_dir)
        else:
            tb_logger = None
    except:
        logger.warning("Tensorboard is not available.")
        tb_logger = None

    return tb_logger



import os
import json
import shutil
import faiss
from tqdm import tqdm


def get_file_size_gb(file_path):
    """Get file size in GB"""
    return os.path.getsize(file_path) / (1024 * 1024 * 1024)

def split_large_file(file_path, output_dir, chunk_size_gb=40):
    """
    Split a large file into smaller chunks.
    For FAISS files, use special handling.
    For other files, use simple binary splitting.
    """
    file_size_gb = get_file_size_gb(file_path)
    if file_size_gb <= chunk_size_gb:
        return False  # No need to split
    
    os.makedirs(output_dir, exist_ok=True)
    file_name = os.path.basename(file_path)
    
    # Special handling for FAISS files
    if file_path.endswith('.faiss'):
        return split_faiss_index(file_path, output_dir, chunk_size_gb)
    
    # For other files, use binary splitting
    chunk_size = int(chunk_size_gb * 1024 * 1024 * 1024)  # Convert to bytes
    with open(file_path, 'rb') as f:
        chunk_num = 0
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break
            
            chunk_path = os.path.join(output_dir, f"{file_name}.chunk_{chunk_num:03d}")
            with open(chunk_path, 'wb') as chunk_file:
                chunk_file.write(chunk)
            chunk_num += 1
    
    # Save metadata
    metadata = {
        'original_file': file_name,
        'n_chunks': chunk_num,
        'chunk_size_gb': chunk_size_gb,
        'total_size_gb': file_size_gb
    }
    
    with open(os.path.join(output_dir, f"{file_name}.metadata.json"), 'w') as f:
        json.dump(metadata, f)
    
    return True

def combine_large_file(input_dir, output_path):
    """
    Combine split file chunks back into a single file.
    For FAISS files, use special handling.
    For other files, use simple binary combining.
    """
    try:
        print(f"Looking for metadata in {input_dir}")
        # Find metadata file - check both naming patterns
        metadata_files = [f for f in os.listdir(input_dir) if f.endswith('.metadata.json') or f == 'index_metadata.json']
        if not metadata_files:
            print(f"No metadata files found in {input_dir}")
            return False
        
        metadata_path = os.path.join(input_dir, metadata_files[0])
        print(f"Found metadata file: {metadata_path}")
        with open(metadata_path, 'r') as f:
            metadata = json.load(f)
        
        # For FAISS files, the metadata might not have original_file
        if 'original_file' not in metadata:
            print("No original_file in metadata, assuming FAISS index")
            return combine_faiss_index(input_dir, output_path)
            
        original_file = metadata['original_file']
        print(f"Original file name: {original_file}")
        
        # Special handling for FAISS files
        if original_file.endswith('.faiss'):
            print("Detected FAISS file, using special handling")
            return combine_faiss_index(input_dir, output_path)
        
        # For other files, use binary combining
        print(f"Combining {metadata['n_chunks']} chunks into {output_path}")
        with open(output_path, 'wb') as outfile:
            for i in range(metadata['n_chunks']):
                chunk_path = os.path.join(input_dir, f"{original_file}.chunk_{i:03d}")
                print(f"Processing chunk {i+1}/{metadata['n_chunks']}: {chunk_path}")
                if not os.path.exists(chunk_path):
                    print(f"Error: Chunk file not found: {chunk_path}")
                    return False
                with open(chunk_path, 'rb') as infile:
                    shutil.copyfileobj(infile, outfile)
        
        print(f"Successfully combined file: {output_path}")
        return True
        
    except Exception as e:
        print(f"Error combining file: {str(e)}")
        import traceback
        print(traceback.format_exc())
        return False

def split_faiss_index(input_path, output_dir, chunk_size_gb=40):
    """Split a large FAISS index into smaller chunks."""
    os.makedirs(output_dir, exist_ok=True)
    
    print(f"Loading index from {input_path}")
    index = faiss.read_index(input_path)
    
    n_vectors = index.ntotal
    dim = index.d
    
    bytes_per_vector = dim * 4
    vectors_per_chunk = int((chunk_size_gb * 1024 * 1024 * 1024) / bytes_per_vector)
    n_chunks = (n_vectors + vectors_per_chunk - 1) // vectors_per_chunk
    
    print(f"Splitting index into {n_chunks} chunks")
    
    for i in tqdm(range(n_chunks)):
        start_idx = i * vectors_per_chunk
        end_idx = min((i + 1) * vectors_per_chunk, n_vectors)
        
        chunk_index = faiss.IndexFlatL2(dim)
        vectors = index.reconstruct_n(start_idx, end_idx - start_idx)
        chunk_index.add(vectors)
        
        chunk_path = os.path.join(output_dir, f"index_chunk_{i:03d}.faiss")
        faiss.write_index(chunk_index, chunk_path)
    
    metadata = {
        'n_chunks': n_chunks,
        'dim': dim,
        'total_vectors': n_vectors,
        'vectors_per_chunk': vectors_per_chunk,
        'index_type': type(index).__name__
    }
    
    with open(os.path.join(output_dir, 'index_metadata.json'), 'w') as f:
        json.dump(metadata, f)
    
    return True

def combine_faiss_index(input_dir, output_path):
    """Combine split FAISS index chunks back into a single index."""
    try:
        print(f"Looking for FAISS metadata in {input_dir}")
        metadata_path = os.path.join(input_dir, 'index_metadata.json')
        if not os.path.exists(metadata_path):
            print(f"FAISS metadata file not found: {metadata_path}")
            return False
            
        with open(metadata_path, 'r') as f:
            metadata = json.load(f)
        
        n_chunks = metadata['n_chunks']
        dim = metadata['dim']
        
        print(f"Combining {n_chunks} FAISS chunks into a single index")
        print(f"Index dimension: {dim}")
        
        combined_index = faiss.IndexFlatL2(dim)
        
        for i in tqdm(range(n_chunks)):
            chunk_path = os.path.join(input_dir, f"index_chunk_{i:03d}.faiss")
            print(f"Processing chunk {i+1}/{n_chunks}: {chunk_path}")
            if not os.path.exists(chunk_path):
                print(f"Error: Chunk file not found: {chunk_path}")
                return False
                
            chunk_index = faiss.read_index(chunk_path)
            vectors = chunk_index.reconstruct_n(0, chunk_index.ntotal)
            combined_index.add(vectors)
        
        print(f"Saving combined index to {output_path}")
        faiss.write_index(combined_index, output_path)
        print("Successfully combined FAISS index")
        return True
        
    except Exception as e:
        print(f"Error combining FAISS index: {str(e)}")
        import traceback
        print(traceback.format_exc())
        return False

def prepare_for_upload(data_dir, chunk_size_gb=40):
    """
    Prepare all large files in the data directory for upload by splitting them if necessary.
    Returns a list of files that were split.
    """
    split_files = []
    for root, _, files in os.walk(data_dir):
        for file in files:
            file_path = os.path.join(root, file)
            if get_file_size_gb(file_path) > chunk_size_gb:
                rel_path = os.path.relpath(file_path, data_dir)
                output_dir = os.path.join(data_dir, f"{rel_path}.split")
                if split_large_file(file_path, output_dir, chunk_size_gb):
                    split_files.append(rel_path)
                    # Remove original file after successful split
                    os.remove(file_path)
    return split_files

def prepare_after_download(data_dir):
    """
    After downloading, combine any split files back into their original form.
    """
    print(f"Scanning {data_dir} for split files...")
    for root, dirs, _ in os.walk(data_dir):
        for dir_name in dirs:
            if dir_name.endswith('.split'):
                split_dir = os.path.join(root, dir_name)
                original_file = dir_name[:-6]  # Remove .split suffix
                output_path = os.path.join(root, original_file)
                print(f"Combining split files from {split_dir} into {output_path}")
                if combine_large_file(split_dir, output_path):
                    print(f"Successfully combined {original_file}")
                    # Remove split directory after successful combination
                    shutil.rmtree(split_dir)
                else:
                    print(f"Warning: Failed to combine {original_file}")
    
    # Special handling for index directory
    index_dir = os.path.join(data_dir, 'index')
    if os.path.exists(index_dir):
        print(f"Checking index directory: {index_dir}")
        index_files = [f for f in os.listdir(index_dir) if f.endswith('.faiss')]
        if not index_files:
            print("No FAISS index found, checking for split files...")
            split_dirs = [d for d in os.listdir(index_dir) if d.endswith('.split')]
            for split_dir in split_dirs:
                split_path = os.path.join(index_dir, split_dir)
                print(f"\nContents of split directory {split_path}:")
                try:
                    files = os.listdir(split_path)
                    print("Files found:")
                    for f in files:
                        file_path = os.path.join(split_path, f)
                        size = os.path.getsize(file_path) / (1024 * 1024)  # Size in MB
                        print(f"  - {f} ({size:.2f} MB)")
                except Exception as e:
                    print(f"Error listing directory contents: {str(e)}")
                
                output_path = os.path.join(index_dir, 'index_IVFFlat.100000.768.2048.faiss')
                print(f"Combining index from {split_path} into {output_path}")
                if combine_large_file(split_path, output_path):
                    print("Successfully combined index")
                    shutil.rmtree(split_path)
                else:
                    print("Warning: Failed to combine index")

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description='Handle large files for Hugging Face upload/download')
    parser.add_argument('--mode', choices=['split', 'combine'], required=True, help='Mode: split for upload or combine after download')
    parser.add_argument('--data_dir', required=True, help='Directory containing the data files')
    parser.add_argument('--chunk_size_gb', type=float, default=40, help='Maximum size of each chunk in GB')
    
    args = parser.parse_args()
    
    if args.mode == 'split':
        split_files = prepare_for_upload(args.data_dir, args.chunk_size_gb)
        print(f"Split {len(split_files)} files: {split_files}")
    else:
        prepare_after_download(args.data_dir)
        print("Combined all split files") 


import os
import re
import logging
import numpy as np
import torch

from .flat import FlatIndexer
from .ivf_flat import IVFFlatIndexer
from .ivf_pq import IVFPQIndexer


class Indexer(object):
    
    def __init__(self, cfg):
        self.cfg = cfg
        self.index_type = cfg.index_type
        
        data_root = os.path.expanduser(self.cfg.data_root)
        
        index_dir = os.path.join(data_root, cfg.domain_name, 'index')
        index_files = [f for f in os.listdir(index_dir) if f.endswith('.faiss')]
        if len(index_files) != 1:
            raise ValueError(f"Expected exactly one .faiss file in {index_dir}, found {len(index_files)}")
        index_path = os.path.join(index_dir, index_files[0])
        meta_file = os.path.join(index_dir, index_files[0].replace('.faiss', '.faiss.meta'))
        assert os.path.exists(index_path) and os.path.exists(meta_file)
        
        passage_dir = os.path.join(data_root, cfg.domain_name, 'passages')
        embedding_paths = None  # assume index is already built
        trained_index_path = None
        pos_map_save_path = os.path.join(data_root, cfg.domain_name, 'passage_pos_id_map.pkl')
        
        if self.index_type == "Flat":
            self.datastore = FlatIndexer(
                embed_paths=embedding_paths,
                index_path=index_path,
                meta_file=meta_file,
                passage_dir=passage_dir,
                pos_map_save_path=pos_map_save_path,
                dimension=None,
            )
        elif self.index_type == "IVFFlat":
            self.datastore = IVFFlatIndexer(
                embed_paths=embedding_paths,
                index_path=index_path,
                meta_file=meta_file,
                trained_index_path=trained_index_path,
                passage_dir=passage_dir,
                pos_map_save_path=pos_map_save_path,
                sample_train_size=None,
                prev_index_path=None,
                dimension=None,
                ncentroids=None,
                probe=self.cfg.nprobe,
            )
        elif self.index_type == "IVFPQ":
            self.datastore = IVFPQIndexer(
                embed_paths=embedding_paths,
                index_path=index_path,
                meta_file=meta_file,
                trained_index_path=trained_index_path,
                passage_dir=passage_dir,
                pos_map_save_path=pos_map_save_path,
                sample_train_size=None,
                prev_index_path=None,
                dimension=None,
                ncentroids=None,
                probe=self.cfg.nprobe,
                n_subquantizers=None,
                code_size=None,
            )
        else:
            raise NotImplementedError
        
        
    def search(self, query_embs, k=5):
        all_scores, all_passages, db_ids = self.datastore.search(query_embs, k)
        return all_scores, all_passages, db_ids
    



from setuptools import setup, find_namespace_packages

setup(
    name="massive_serve",
    version="0.1.13",
    packages=find_namespace_packages(include=['massive_serve*']),
    package_data={
        'massive_serve': [
            'api/**/*',
            'src/**/*',
            'contriever/**/*',
            'utils/**/*',
            'contriever/src/**/*',
        ],
    },
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "click",  # Required for CLI
        "faiss-cpu",  # or faiss-gpu if you're using GPU
        "tqdm",
        "flask",
        "flask-cors",
        "torch",
        "transformers",
        "numpy",
        "sentence-transformers",
    ],
    extras_require={
        'st': ['sentence-transformers'],
    },
    entry_points={
        'console_scripts': [
            'massive-serve=massive_serve.cli:cli',
        ],
    },
    author="Rulin Shao",
    author_email="rulins@cs.washington.edu",
    description="A package for massive serving",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/RulinShao/massive-serve",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
) 



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import torch
import torch.distributed as dist


class Gather(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x: torch.tensor):
        output = [torch.zeros_like(x) for _ in range(dist.get_world_size())]
        dist.all_gather(output, x)
        return tuple(output)

    @staticmethod
    def backward(ctx, *grads):
        all_gradients = torch.stack(grads)
        dist.all_reduce(all_gradients)
        return all_gradients[dist.get_rank()]


def gather(x: torch.tensor):
    if not dist.is_initialized():
        return x
    x_gather = Gather.apply(x)
    x_gather = torch.cat(x_gather, dim=0)
    return x_gather


@torch.no_grad()
def gather_nograd(x: torch.tensor):
    if not dist.is_initialized():
        return x
    x_gather = [torch.ones_like(x) for _ in range(dist.get_world_size())]
    dist.all_gather(x_gather, x, async_op=False)

    x_gather = torch.cat(x_gather, dim=0)
    return x_gather


@torch.no_grad()
def varsize_gather_nograd(x: torch.Tensor):
    """gather tensors of different sizes along the first dimension"""
    if not dist.is_initialized():
        return x

    # determine max size
    size = torch.tensor([x.shape[0]], device=x.device, dtype=torch.int)
    allsizes = [torch.zeros_like(size) for _ in range(dist.get_world_size())]
    dist.all_gather(allsizes, size)
    max_size = max([size.cpu().max() for size in allsizes])

    padded = torch.empty(max_size, *x.shape[1:], dtype=x.dtype, device=x.device)
    padded[: x.shape[0]] = x
    output = [torch.zeros_like(padded) for _ in range(dist.get_world_size())]
    dist.all_gather(output, padded)

    output = [tensor[: allsizes[k]] for k, tensor in enumerate(output)]
    output = torch.cat(output, dim=0)

    return output


@torch.no_grad()
def get_varsize(x: torch.Tensor):
    """gather tensors of different sizes along the first dimension"""
    if not dist.is_initialized():
        return [x.shape[0]]

    # determine max size
    size = torch.tensor([x.shape[0]], device=x.device, dtype=torch.int)
    allsizes = [torch.zeros_like(size) for _ in range(dist.get_world_size())]
    dist.all_gather(allsizes, size)
    allsizes = torch.cat(allsizes)
    return allsizes


def get_rank():
    if not dist.is_available():
        return 0
    if not dist.is_initialized():
        return 0
    return dist.get_rank()


def is_main():
    return get_rank() == 0


def get_world_size():
    if not dist.is_initialized():
        return 1
    else:
        return dist.get_world_size()


def barrier():
    if dist.is_initialized():
        dist.barrier()


def average_main(x):
    if not dist.is_initialized():
        return x
    if dist.is_initialized() and dist.get_world_size() > 1:
        dist.reduce(x, 0, op=dist.ReduceOp.SUM)
        if is_main():
            x = x / dist.get_world_size()
    return x


def sum_main(x):
    if not dist.is_initialized():
        return x
    if dist.is_initialized() and dist.get_world_size() > 1:
        dist.reduce(x, 0, op=dist.ReduceOp.SUM)
    return x


def weighted_average(x, count):
    if not dist.is_initialized():
        if isinstance(x, torch.Tensor):
            x = x.item()
        return x, count
    t_loss = torch.tensor([x * count]).cuda()
    t_total = torch.tensor([count]).cuda()
    t_loss = sum_main(t_loss)
    t_total = sum_main(t_total)
    return (t_loss / t_total).item(), t_total.item()



import os
import sys
import json
import pickle as pkl
import logging
import time
import copy
import random
from tqdm import tqdm
import re
import pdb
import string
from collections import Counter

import torch


os.environ["TOKENIZERS_PARALLELISM"] = "true"


device = 'cuda' if torch.cuda.is_available()  else 'cpu'


def embed_queries(cfg, queries, model, tokenizer, model_name_or_path):
    if "sentence-transformers" in model_name_or_path or "e5" in model_name_or_path:
        all_question = []
        for k, q in enumerate(queries):
            all_question.append(q)
        
        embeddings = model.encode(all_question, batch_size=min(128, cfg.per_gpu_batch_size))  # sentence-transformer has extra memory overhead and can only support a smaller batch size
    
    else:
        model.eval()
        embeddings, batch_question = [], []
        with torch.no_grad():

            for k, q in tqdm(enumerate(queries)):
                batch_question.append(q)

                if len(batch_question) == cfg.per_gpu_batch_size or k == len(queries) - 1:
                    
                    if "drama" in model_name_or_path:
                        output = model.encode_queries(batch_question, batch_question, dim=768)  # TODO: change this to align with index.projection_size
                    elif "ReasonIR" in model_name_or_path or "GRIT" in model_name_or_path:
                        output = model.encode(batch_question, instruction="", batch_size=cfg.per_gpu_batch_size)
                        output = torch.tensor(output, device='cpu')
                    else:
                        encoded_batch = tokenizer.batch_encode_plus(
                            batch_question,
                            return_tensors="pt",
                            max_length=cfg.question_maxlength,
                            padding=True,
                            truncation=True,
                        )

                        encoded_batch = {k: v.to(device) for k, v in encoded_batch.items()}
                        output = model(**encoded_batch)
                        if "contriever" not in model_name_or_path:
                            output = output.last_hidden_state[:, 0, :]
                    
                    embeddings.append(output.cpu())

                    batch_question = []

        embeddings = torch.cat(embeddings, dim=0).numpy()
    
    print(f"Questions embeddings shape: {embeddings.shape}")

    return embeddings



import os
import time
import json
from massive_serve.contriever import Contriever, load_retriever
from transformers import AutoModel
import torch

from massive_serve.src.search import embed_queries
from massive_serve.src.indicies.base import Indexer


device = 'cuda' if torch.cuda.is_available()  else 'cpu'

class DatastoreAPI():
    def __init__(self, cfg) -> None:
        self.domain_name = cfg.domain_name
        self._index = Indexer(cfg)
        self.index = self._index.datastore
        
        model_name_or_path = cfg.query_encoder
        tokenizer_name_or_path = cfg.query_tokenizer
        if "contriever" in model_name_or_path:
            query_encoder, query_tokenizer, _ = load_retriever(model_name_or_path)
        elif "dragon" in model_name_or_path or "drama" in model_name_or_path or "ReasonIR" in model_name_or_path:
            query_encoder = AutoModel.from_pretrained(model_name_or_path, trust_remote_code=True)
        elif "sentence-transformers" in model_name_or_path or "e5" in model_name_or_path:
            try:
                from sentence_transformers import SentenceTransformer
            except:
                print(f"{model_name_or_path} requires sentence-transformers. Please install it with `pip install sentence-transformers`.")
                raise AttributeError
            query_tokenizer = None
            query_encoder = SentenceTransformer(model_name_or_path)
        elif "GRIT" in model_name_or_path:
            from gritlm import GritLM
            query_tokenizer = None
            query_encoder = GritLM(model_name_or_path, torch_dtype="auto", mode="embedding")
        else:
            print(f"{model_name_or_path} is not supported!")
            raise AttributeError
        self.query_encoder = query_encoder
        self.query_tokenizer = query_tokenizer
        self.query_encoder = self.query_encoder.to(device)
        
        self.cfg = cfg
    
    def search(self, query, n_docs=3):
        query_embedding = self.embed_query(query)
        searched_scores, searched_passages, db_ids  = self.index.search(query_embedding, n_docs)
        results = {'scores': searched_scores, 'passages': searched_passages, 'IDs': db_ids}
        return results
    
    def embed_query(self, query):
        if isinstance(query, str):
            query_embedding = embed_queries(self.cfg, [query], self.query_encoder, self.query_tokenizer, self.cfg.query_encoder)
        elif isinstance(query, list):
            query_embedding = embed_queries(self.cfg, query, self.query_encoder, self.query_tokenizer, self.cfg.query_encoder)
        else:
            raise AttributeError(f"Query is not a string nor list!")
        return query_embedding
    

def get_datastore(cfg):
    ds = DatastoreAPI(cfg)
    test_search(ds)
    return ds


def main(cfg):
    get_datastore(cfg, 0)


def test_search(ds):
    query = 'when was the last time anyone was on the moon?'  # 'scores': array([[44.3889  , 44.770973, 45.956238]], dtype=float32), 'IDs': [[[5, 45516], [6, 2218998], [5, 897337]]
    query2 = "who wrote he ain't heavy he's my brother lyrics?"  # 'scores': array([[33.60194 , 41.798004, 43.465225]], dtype=float32) 'IDs': [[[2, 361677], [5, 1717105], [2, 361675]]]
    search_results = ds.search(query, 1)
    print(search_results)



if __name__ == '__main__':
    main()
    
    